/*  mgmtmsg.cpp */

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

#ifdef WIN32
#include <windows.h>
#include "largeint.h"
#include <wchar.h>
#endif

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "BipSampPluginUtil.h"
#include <BipCpi.h>
#include <BipCni.h>

#include "common.h"
#include "parsubs.h"
#include "trace.h"
#include "metadata.h"
#include "fixparser.h"
#include "fixsub.h"

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#define FLUSH		"FLUSH   "
#define FLUSHALL	"FLUSHALL"
#define STATS		"STATS   "
#define TRACEON		"TRACEON "
#define TRACEOFF	"TRACEOFF"
#define GETLEVEL	"GETLEVEL"
#define FLUSH_SIZE	8

#ifdef _DEBUG
static char Level[]="mgmtmsg.cpp V1.0 Debug version ("__DATE__" "__TIME__")";
#else
static char Level[]="mgmtmsg.cpp V1.0 Release version ("__DATE__" "__TIME__")";
#endif

#ifdef _DEBUG
	extern int	debug;
	extern int	traceParser;
	extern int	traceWrite;
	extern int	traceModule;
	extern int	traceMgmt;
#endif

	extern int	parserStartHigh;	// Time parser was started (high word)
	extern int	parserStartLow;		// Time parser was started (low word)
#ifdef WIN32
	extern __int64 totalParseTime;	// Total parsing time in microseconds
	extern __int64 totalWriteTime;	// Total time to build a buffer in microseconds
	extern __int64 maxParseTime;	// Maximum time to parse a message in microseconds
	extern __int64 minParseTime;	// Minimum time to parse a message in microseconds
	extern __int64 maxWriteTime;	// Maximum time to build a buffer in microseconds
	extern __int64 minWriteTime;	// Minimum time to build a buffer in microseconds
#else
	extern long long totalParseTime; // Total parsing time in microseconds
	extern long long totalWriteTime; // Total time to build a buffer in microseconds
	extern long long maxParseTime;	// Maximum time to parse a message in microseconds
	extern long long minParseTime;	// Minimum time to parse a message in microseconds
	extern long long maxWriteTime;	// Maximum time to build a buffer in microseconds
	extern long long minWriteTime;	// Minimum time to build a buffer in microseconds
#endif
	extern int	totalBytes;			// Number of bytes in all messages processed
	extern int	maxMessage;			// Largest message processed
	extern int	minMessage;			// Smallest message processed
	extern int	totalBytes;			// Number of bytes in all messages processed
	extern int	writeCount;			// Number of cpiWriteBuffer calls processed
	extern int	totalWritten;		// Number of bytes written
	extern int	maxWriteLength;		// Largest message written
	extern int	minWriteLength;		// Smallest message written

	extern int	messageCount;		// Number of messages processed    
	extern int	badMessageCount;	// Number of input messages rejected    
	extern int	badOutputCount;		// Number of output messages rejected    
	extern int	inTagCount;			// Number of tags parsed in input messages
	extern int	outTagCount;		// Number of tags created in output messages

	extern int	useCounter;			// Counter for least recently used meta data files
	extern int	filesOpened;		// files added to cache
	extern int	filesRemoved;		// files removed from cache
	extern int	filesMax;			// Maximum number of files in cache

/******************************************************************************/
/*                                                                            */
/* Routine to process a cache flush management message.                    .  */
/*                                                                            */
/******************************************************************************/

void processFlushMsg(int ver, int mod, unsigned char * fileCodes, int length)

{
	int	i;
	int	j;
	int	count;
	char msgType[FLUSH_SIZE + 1];

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
	    Log("PLUGIN: -> processFlushMsg()\n");
	}
#endif

	if (length >= FLUSH_SIZE)
	{
		// calculate the number of entries in the message
		count = length / FLUSH_SIZE;

		i = 0;
		while (i < count)
		{
			// get the message type name as a string
			j = 0;
			while ((j < FLUSH_SIZE) && (fileCodes[j] > ' '))
			{
				msgType[j] = fileCodes[j];
				j++;
			}

			// terminate the string
			msgType[j] = 0;

			removeMetaDataEntry(msgType);

			// move on to the next entry
			fileCodes += FLUSH_SIZE;
			i++;
		}
	}
	else
	{
		// Invalid message, no entries to flush
#ifdef _DEBUG
		if (traceModule)
		{
			Log("No entries - length(%d)\n", length);
		}
#endif
	}

#ifdef _DEBUG
	if (traceModule)
	{
	    Log("PLUGIN: <- processFlushMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a cache flush all management message.                .  */
/*                                                                            */
/******************************************************************************/

void processFlushAllMsg()

{
	// Flush all the metadata entries
	removeAllMetaData();
}

/******************************************************************************/
/*                                                                            */
/* Routine to add a name element to the parse tree.                        .  */
/*                                                                            */
/******************************************************************************/

CciElement * newNameElement(CciParser* parser, const char* name)

{
	int			rc;
	CciElement* newElement;
	CciChar		Uname[64];
	CciSize		size;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> newNameElement - creating name element for (%s)\n", name);
	}
#endif

	// translate the input name to Unicode
	size = cciMbsToUcs(&rc, name, Uname, sizeof(Uname) - 1, DEF_CODEPAGE);

	// Create a new name element for the tag 
	newElement = cpiCreateAndInitializeElement(&rc, 
											   parser, 
											   CCI_ELEMENT_TYPE_NAME, 
											   Uname, 
											   COMPLETE_BIT,
											   COMPLETE_BIT);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- newNameElement - created name element 0x%x\n", newElement);
	}
#endif

	return newElement;
}

///////////////////////////////////////////////////
//
// Routine to add a statistics element, with
// the given name and value under the
// specified parent element.
//
///////////////////////////////////////////////////

void addStatisticElement(CciParser  * parser, 
						 const char * valueChar, 
						 const char * name, 
						 CciElement * parent)

{
	int			rc;
	int			size;
	CciElement* newElement;
	CciChar		UvalueStr[256];
	CciChar		Uname[64];

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> addStatElement - element for (%s)\n", name);
	}
#endif

	// translate the input name to Unicode
	size = cciMbsToUcs(&rc, name, Uname, sizeof(Uname) - 1, DEF_CODEPAGE);

	// Create a new element for this tag 
	// Create a new name element for the tag 
	newElement = cpiCreateAndInitializeElement(&rc, 
											   parser, 
											   CCI_ELEMENT_TYPE_NAME_VALUE, 
											   Uname, 
											   COMPLETE_BIT,
											   COMPLETE_BIT);

	// Copy the text into a wide ImbString 
	UvalueStr[0] = 0;
	size = cciMbsToUcs(&rc, valueChar, UvalueStr, sizeof(UvalueStr) - 1, DEF_CODEPAGE);


	// Set the element value
	cpiSetElementCharacterValue(&rc, newElement, UvalueStr, size);

	// Mark the element complete and add it to the tree
	cpiAddAsLastChild(&rc, parent, newElement);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- addStatElement - element 0x%x created as child of 0x%x for (%S)\n", newElement, parent, name);
	}
#endif
}

///////////////////////////////////////////////////
//
// Routine to add a statistics element, with
// the given name and integer value under the
// specified parent element.
//
///////////////////////////////////////////////////

void addStatElement(CciParser  * parser, 
					const int value, 
					const char * name, 
					CciElement * parent)

{
	char		valueChar[32];

	// now, add the value element
	sprintf(valueChar, "%d", value);

	// create the element
	addStatisticElement(parser, valueChar, name, parent);
}

///////////////////////////////////////////////////
//
// Routine to add a statistics element, with
// the given name and integer value under the
// specified parent element.  The integer is
// assumed to be in microseconds, with the result
// to be displayed as nnn.nn milliseconds.
//
///////////////////////////////////////////////////

#ifdef WIN32
void addStatTimeElement(CciParser  * parser, 
						const __int64 value, 
						const char * name, 
						CciElement * parent)
#else
void addStatTimeElement(CciParser  * parser, 
						const long long value, 
						const char * name, 
						CciElement * parent)
#endif

{
#ifdef WIN32
	__int64		tempValue;
#else
	long long	tempValue;
#endif

	char		valueChar[32];
	char		valueChar2[4];

	// now, add the value element
#ifdef WIN32
	sprintf(valueChar, "%I64d", value / 1000);
#else
	sprintf(valueChar, "%lld", value / 1000);
#endif

	// get the tenths and hundreths of a millisecond
	tempValue = (value / 10) % 100;

	// get the tenths and hundredths as a character value */
#ifdef WIN32
	sprintf(valueChar2, "%2.2I64d", tempValue);
#else
	sprintf(valueChar2, "%2.2lld", tempValue);
#endif

	if (' ' == valueChar2[0])
	{
		valueChar2[0] = '0';
	}

	// append the tenths and hundreths
	strcat(valueChar, ".");
	strcat(valueChar, valueChar2);

	// create the element
	addStatisticElement(parser, valueChar, name, parent);
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a statistics management message.                     .  */
/*                                                                            */
/*  Build a parse tree of the following format (as if in XML)                 */
/*   <FIX>                                                                    */
/*    <statistics>                                                            */
/*     <cache>                                                                */
/*      <filesOpened>nnn</filesOpened>                                        */
/*      <filesRemoved>nnn</filesRemoved>                                      */
/*      <filesMax>nnn</filesMax>                                              */
/*      <useCounter>nnn</useCounter>                                          */
/*     </cache>                                                               */
/*     <messages>                                                             */
/*      <messageCount>nnn</messageCount>                                      */
/*      <inTagCount>nnn</inTagCount>                                          */
/*      <outTagCount>nnn</outTagCount>                                        */
/*      <invalidMsgCount>nnn</invalidMsgCount>                                */
/*      <invalidOutputCount>nnn</invalidOutputCount>                          */
/*      <writeBufferCount>nnn</writeBufferCount>                              */
/*      <maxMessageSize>nnn</maxMessageSize>                                  */
/*      <minMessageSize>nnn</minMessageSize>                                  */
/*      <averageMessageSize>nnn</averageMessageSize>                          */
/*      <averageParseTime>nnn.nn</averageParseTime>                           */
/*      <maxParseTime>nnn.nn</maxParseTime>                                   */
/*      <minParseTime>nnn.nn</minParseTime>                                   */
/*      <averageOutputSize>nnn</averageOutputSize>                            */
/*      <maxOutputSize>nnn</maxOutputSize>                                    */
/*      <minOutputSize>nnn</minOutputSize>                                    */
/*      <averageWriteTime>nnn.nn</averageWriteTime>                           */
/*      <maxWriteTime>nnn.nn</maxWriteTime>                                   */
/*      <minWriteTime>nnn.nn</minWriteTime>                                   */
/*     </messages>                                                            */
/*    </statistics>                                                           */
/*   </FIX>                                                                   */
/*                                                                            */
/******************************************************************************/

void processStatsMsg(CciParser* parser, int ver, int mod, unsigned char * statCodes)

{
	int			rc;
	int			average;
	int			minSize;

#ifdef WIN32
	__int64		minTime;
	__int64		averageTime=0;
#else
	long long	minTime;
	long long	averageTime=0;
#endif

	CciElement* root;
	CciElement* nameElement;
	CciElement* statisticsElement;
	CciElement* cacheElement;
	CciElement* messagesElement;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> processStatsMsg() ver %d mod %d\n", ver, mod);
	}
#endif

	// Get a pointer to the root element for the parser
	root = cpiRootElement(&rc, parser);

	// Make sure the top level is set to complete
	cpiSetElementCompletePrevious(&rc, root, 1);
	cpiSetElementCompleteNext(&rc, root, 1);

	// create the statistics element
	// Create a new name element for the tag 
	nameElement = newNameElement(parser, "statistics");
	cpiAddAsFirstChild(&rc, root, nameElement);
	statisticsElement = nameElement;

	// create the cache element
	// Create a new name element for the tag 
	nameElement = newNameElement(parser, "cache");
	cpiAddAsFirstChild(&rc, statisticsElement, nameElement);
	cacheElement = nameElement;

	// create the messages element under statistics
	messagesElement = newNameElement(parser, "messages");
	cpiAddAfter(&rc, cacheElement, messagesElement);

	addStatElement(parser, filesOpened, "filesOpened", cacheElement);
	addStatElement(parser, filesRemoved, "filesRemoved", cacheElement);
	addStatElement(parser, filesMax, "filesMax", cacheElement);
	addStatElement(parser, useCounter, "useCounter", cacheElement);

	addStatElement(parser, messageCount, "inMessageCount", messagesElement);
	addStatElement(parser, inTagCount, "inTagCount", messagesElement);
	addStatElement(parser, badMessageCount, "invalidMsgCount", messagesElement);

	/* get the average input message size */
	average = 0;
	if ((messageCount - badMessageCount) > 0)
	{
		average = totalBytes / (messageCount - badMessageCount);
	}

	/* get the minimum input message size or 0 if no messages have been parsed */
	if (-1 == minMessage)
	{
		minSize = 0;
	}
	else
	{
		minSize = minMessage;
	}

	addStatElement(parser, average, "averageMessageSize", messagesElement);
	addStatElement(parser, maxMessage, "maxMessageSize", messagesElement);
	addStatElement(parser, minSize, "minMessageSize", messagesElement);

	// get the average parse time
	averageTime = 0;
	if ((messageCount - badMessageCount) > 0)
	{
		averageTime = totalParseTime  / (messageCount - badMessageCount);
	}

	/* get the minimum parse time or 0 if no messages have been parsed */
	if (-1 == minParseTime)
	{
		minTime = 0;
	}
	else
	{
		minTime = minParseTime;
	}

	addStatTimeElement(parser, averageTime, "averageParseTime", messagesElement);
	addStatTimeElement(parser, maxParseTime, "maxParseTime", messagesElement);
	addStatTimeElement(parser, minTime, "minParseTime", messagesElement);

	/* get the average output message size */
	average = 0;
	if ((writeCount - badOutputCount) > 0)
	{
		average = totalWritten / (writeCount - badOutputCount);
	}

	/* get the minimum output message size or 0 if no messages have been written */
	if (-1 == minWriteLength)
	{
		minSize = 0;
	}
	else
	{
		minSize = minWriteLength;
	}

	addStatElement(parser, writeCount, "writeCount", messagesElement);
	addStatElement(parser, outTagCount, "outTagCount", messagesElement);
	addStatElement(parser, badOutputCount, "invalidOutputCount", messagesElement);
	addStatElement(parser, average, "averageOutputSize", messagesElement);
	addStatElement(parser, maxWriteLength, "maxOutputSize", messagesElement);
	addStatElement(parser, minSize, "minOutputSize", messagesElement);

	// calculate the average time to create an output message
	averageTime = 0;
	if ((writeCount - badOutputCount) > 0)
	{
		averageTime = totalWriteTime / (writeCount - badOutputCount);
	}

	/* get the minimum write time or 0 if no messages have been written */
	if (-1 == minWriteTime)
	{
		minTime = 0;
	}
	else
	{
		minTime = minWriteTime;
	}

	addStatTimeElement(parser, averageTime, "averageWriteTime", messagesElement);
	addStatTimeElement(parser, maxWriteTime, "maxWriteTime", messagesElement);
	addStatTimeElement(parser, minTime, "minWriteTime", messagesElement);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- processStatsMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a request to turn tracing on.                        .  */
/*                                                                            */
/******************************************************************************/

void processTraceOnMsg(int ver, int mod, int length, unsigned char * traceTypes)

{
#ifdef _DEBUG
	int	count=0;

	// Tell what we are doing
	if (traceModule)
	{
		Log("PLUGIN: -> processTraceOnMsg() ver %d mod %d\n", ver, mod);
	}

	/* calculate the number of requests in the message */
	if (length > 0)
	{
		count = length / 8;
	}

	/* Turn on the requested traces */
	processTraceMessage(traceTypes, count, 1);

	if (traceModule)
	{
		Log("PLUGIN: <- processTraceOnMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a request to turn tracing off.                       .  */
/*                                                                            */
/******************************************************************************/

void processTraceOffMsg()

{
#ifdef _DEBUG
	unsigned char	tempMsg[] = TRACEOFF;

	// Tell what we are doing
	if (traceModule)
	{
		Log("PLUGIN: -> processTraceOffMsg()\n");
		Log("Closing trace file\n");
	}

	processTraceMessage(tempMsg, 1, 0);

	if (traceModule)
	{
		Log("PLUGIN: <- processTraceOnMsg()\n");
	}
#endif
}

void addCharacterElement(CciParser* parser, 
						 CciElement *parent, 
						 const char * elementValue,
						 const char * elementName)

{
	int			rc;
	int			size;
	CciChar		workValue[1024];
	CciChar		Uname[64];
	CciElement*	newElement;

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> addCharacterElement - element %s as child of 0x%x\n", elementName, parent);
	}
#endif

	// translate the input name to Unicode
	size = cciMbsToUcs(&rc, elementName, Uname, sizeof(Uname) - 1, DEF_CODEPAGE);

	// Create a new name element for the tag 
	newElement = cpiCreateAndInitializeElement(&rc, 
											   parser, 
											   CCI_ELEMENT_TYPE_NAME_VALUE,
											   Uname,
											   COMPLETE_BIT,
											   COMPLETE_BIT);

	// Copy the text into a wide ImbString 
	workValue[0] = 0;
	size = cciMbsToUcs(&rc, elementValue, workValue, sizeof(workValue) - 1, DEF_CODEPAGE);

	// Set the element value
	cpiSetElementCharacterValue(&rc, newElement, workValue, size);

	// add the element to the tree
	cpiAddAsLastChild(&rc, parent, newElement);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- addCharacterElement - element %S created with value (%s)\n", elementName, elementValue);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a request to get the module levels.                  .  */
/*                                                                            */
/*  Build a parse tree of the following format (as if in XML)                 */
/*   <FIX>                                                                    */
/*    <ModuleLevels>                                                          */
/*     <Fixparser>nnn</Fixparser>                                             */
/*     <Fixsub>nnn</Fixsub>                                                   */
/*     <Metadata>nnn</Metadata>                                               */
/*     <Mgmtmsg>nnn</Mgmtmsg>                                                 */
/*     <Parsubs>nnn</Parsubs>                                                 */
/*     <Trace>nnn</Trace>                                                     */
/*    </ModuleLevels>                                                         */
/*   </FIX>                                                                   */
/******************************************************************************/

void processGetlevelMsg(CciParser* parser)

{
	int			rc;
	const char*	moduleLevel;
	CciElement* root;
	CciElement* levelElement;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> processGetlevelMsg()\n");
	}
#endif

	// Get a pointer to the root element for the parser
	root = cpiRootElement(&rc, parser);

	// Make sure the top level is set to complete
	cpiSetElementCompletePrevious(&rc, root, 1);
	cpiSetElementCompleteNext(&rc, root, 1);

	// create the module level element
	// Create a new name element for the tag 
	levelElement = newNameElement(parser, "ModuleLevels");
	cpiAddAsFirstChild(&rc, root, levelElement);

	moduleLevel = getFixParserLevel();
	addCharacterElement(parser, levelElement, moduleLevel, "Fixparser");

	moduleLevel = getFixSubLevel();
	addCharacterElement(parser, levelElement, moduleLevel, "Fixsub");

	moduleLevel = getMetadataLevel();
	addCharacterElement(parser, levelElement, moduleLevel, "Metadata");

	addCharacterElement(parser, levelElement, Level, "MgmtMsg");

	moduleLevel = getParsubsLevel();
	addCharacterElement(parser, levelElement, moduleLevel, "Parsubs");

	moduleLevel = getTraceLevel();
	addCharacterElement(parser, levelElement, moduleLevel, "Trace");

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- processGetlevelMsg()\n");
	}
#endif
}
/******************************************************************************/
/*                                                                            */
/* Routine to parse management message and perform the necessary action.   .  */
/*                                                                            */
/******************************************************************************/

int parseMgmtMsg(CciParser* parser, PARSER_CONTEXT_ST* context)

{
	int		rc=0;
	int		ver;
	int		mod;
	int		size;
	char	msgType[9];
	char	version[5];
	char	modLevel[5];
	CciElement*	body;
	unsigned char	*buffer=NULL;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> parseMgmtMsg() length %d\n", context->iSize);
	}
#endif

	/* get the input data and translate it to the local code page */
	buffer = (unsigned char *)malloc(context->iSize + 1);
	size = cciUcsToMbs(&rc, context->iBuffer, (char *)buffer, context->iSize, DEF_CODEPAGE);

	// get the type of management message
	memcpy(msgType, buffer + 8, 8);
	msgType[8] = 0;

	// get the version number and modification level
	memcpy(version, buffer, 4);
	version[4] = 0;
	ver = atoi(version);
	memcpy(modLevel, buffer + 4, 4);
	modLevel[4] = 0;
	mod = atoi(modLevel);

	if ((strcmp(version, "0000") == 0) && (strcmp(modLevel, "0000") == 0)) 
	{
		// Tell what we are doing
#ifdef _DEBUG
		if (traceMgmt)
		{
			Log("parsing management message type(%s) version(%s) modification level(%s)\n", msgType, version, modLevel);
		}
#endif

		if (strcmp(msgType, FLUSH) == 0)
		{
			processFlushMsg(ver, mod, buffer + 16, context->iSize - 16);
		}

		if (strcmp(msgType, FLUSHALL) == 0)
		{
			processFlushAllMsg();
		}

		if (strcmp(msgType, STATS) == 0)
		{
			processStatsMsg(parser, ver, mod, buffer + 16);
		}

		if (strcmp(msgType, TRACEON) == 0)
		{
			processTraceOnMsg(ver, mod, context->iSize - 16, buffer + 16);
		}

		if (strcmp(msgType, TRACEOFF) == 0)
		{
			processTraceOffMsg();
		}

		if (strcmp(msgType, GETLEVEL) == 0)
		{
			processGetlevelMsg(parser);
		}
	}
	else
	{
		// wrong version or modification level
#ifdef _DEBUG
		if (traceMgmt)
		{
			Log("Wrong version(%s) or modification level(%s)\n", version, modLevel);
		}
#endif
	}

	// finally, we must mark the body element complete, in case
	// some node tries to use this message
	body = cpiRootElement(&rc, parser);
	if (body != CCI_NULL_ADDR)
	{
		cpiSetElementCompleteNext(&rc, body, 1);
		cpiSetElementCompletePrevious(&rc, body, 1);
	}

	/* release the buffer storage */
	free(buffer);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- parseMgmtMsg()\n");
	}
#endif

	return rc;
}

//////////////////////
// End of file
//////////////////////